from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.main import MainHandler
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.image import ImageHandler
from RGBMatrixEmulator.adapters.browser_adapter.request_handlers.image_web_socket import ImageWebSocketHandler


import importlib, json

from RGBMatrixEmulator.logger import Logger

adapters = [
    {
        'path': 'RGBMatrixEmulator.adapters.browser_adapter.adapter',
        'class': 'BrowserAdapter',
        'type': 'browser'
    },
    {
        'path': 'RGBMatrixEmulator.adapters.pygame_adapter',
        'class': 'PygameAdapter',
        'type': 'pygame'
    },
    {
        'path': 'RGBMatrixEmulator.adapters.sixel_adapter',
        'class': 'SixelAdapter',
        'type': 'sixel'
    },
    {
        'path': 'RGBMatrixEmulator.adapters.terminal_adapter',
        'class': 'TerminalAdapter',
        'type': 'terminal'
    },
    {
        'path': 'RGBMatrixEmulator.adapters.tkinter_adapter',
        'class': 'TkinterAdapter',
        'type': 'tkinter'
    },
    {
        'path': 'RGBMatrixEmulator.adapters.turtle_adapter',
        'class': 'TurtleAdapter',
        'type': 'turtle'
    }
]

ADAPTER_TYPES = {}

try:
    with open("emulator_config.json") as config_file:
        suppress_adapter_load_errors = json.load(config_file).get("suppress_adapter_load_errors", False)
except:
    suppress_adapter_load_errors = False

for adapter in adapters:
    package_path = adapter.get('path')
    adapter_class = adapter.get('class')
    adapter_name = adapter.get('type')
    try:
        package = importlib.import_module(package_path)
        adapter = getattr(package, adapter_class)

        ADAPTER_TYPES[adapter_name] = adapter
    except Exception as ex:
        if suppress_adapter_load_errors: continue

        Logger.exception(f'''
Failed to load {adapter_class} for "{adapter_name}" display adapter!

If this is not your configured display adapter, the emulator will continue to load.

You can suppress this error in the `emulator_config.json` by adding:

  "suppress_adapter_load_errors": true

'''
        )



import os
import tkinter
import turtle

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.graphics.color import Color
from RGBMatrixEmulator.logger import Logger


class TurtleAdapter(BaseAdapter):
    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__pen = None
        self.__screen = None

    def draw_to_screen(self, pixels):
        self.__pen.clear()

        for row, pixel_row in enumerate(pixels):
            self.__move_pen_to_row_start(row)

            for _col, pixel in enumerate(pixel_row):
                pixel = self.adjust_pixel_brightness(pixel)
                self.__draw_pixel(pixel)
                self.__move_pen_next_pixel()

        self.__screen.update()

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info('Loading {}'.format(self.emulator_details_text()))
        turtle.setup(*self.options.window_size())
        turtle.title(self.emulator_details_text())
        self.__pen = turtle.Turtle(visible = False)
        self.__screen = self.__pen.getscreen()
        self.__set_emulator_icon()
        self.__screen.bgcolor(Color.BLACK())
        turtle.tracer(0, 0)
        turtle.colormode(255)

        self.loaded = True

    def __draw_pixel(self, pixel):
        self.__pen.color(*pixel)
        self.__pen.begin_fill()

        if self.options.pixel_style == 'circle':
            self.__draw_circle_pixel()
        else:
            self.__draw_square_pixel()

        self.__pen.end_fill()

        self.__pen.setheading(0)

    def __draw_square_pixel(self):
        for _ in range(0, 4):
            self.__pen.forward(self.options.pixel_size)
            self.__pen.left(90)

    def __draw_circle_pixel(self):
        self.__pen.pendown()
        self.__pen.dot(self.options.pixel_size)
        self.__pen.penup()

        # Apparently dots cannot overlap, so set movement to the smallest increment possible
        self.__pen.forward(1)

    def __move_pen_next_pixel(self):
        self.__pen.penup()
        self.__pen.forward(self.options.pixel_size)
        self.__pen.pendown()

    def __move_pen_to_row_start(self, row_number):
        self.__reset_pen_position()
        self.__pen.penup()
        self.__pen.setheading(270)
        self.__pen.forward(self.options.pixel_size * row_number)
        self.__pen.setheading(0)
        self.__pen.pendown()

    def __reset_pen_position(self):
        self.__pen.penup()
        self.__pen.goto(self.options.pixel_size / 2 - self.__screen.window_width() / 2, self.__screen.window_height() / 2 - self.options.pixel_size / 2)
        self.__pen.pendown()

    def __set_emulator_icon(self):
        emulator_path = os.path.abspath(os.path.dirname(__file__))
        raw_icon_path = os.path.join(emulator_path, '..', 'icon.png')
        icon_path = os.path.normpath(raw_icon_path)

        icon_image = tkinter.Image("photo", file=icon_path)
        self.__screen._root.iconphoto(True, icon_image)



import tornado.web


class ImageHandler(tornado.web.RequestHandler):
    def get(self):
        self.set_header("Content-type", "image/{}".format(self.adapter.image_format.lower()))
        self.write(self.adapter.image)

    def register_adapter(adapter):
        ImageHandler.adapter = adapter



import io

import numpy as np

from PIL import Image, ImageDraw
from RGBMatrixEmulator.graphics import Color
from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.adapters.browser_adapter.server import Server
from RGBMatrixEmulator.logger import Logger


class BrowserAdapter(BaseAdapter):

    SUPPORTS_ALTERNATE_PIXEL_STYLE = True
    IMAGE_FORMATS = {
        'bmp': 'BMP',
        'jpeg': 'JPEG',
        'png':  'PNG',
        'webp': 'WebP'
    }

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__server = None
        self.image = None
        self.default_image_format = "JPEG"

        image_format = options.browser.image_format
        if image_format.lower() in self.IMAGE_FORMATS:
            self.image_format = self.IMAGE_FORMATS[image_format.lower()]
        else:
            Logger.warning("Invalid browser image format '{}', falling back to '{}'".format(image_format, self.default_image_format))
            self.image_format = self.IMAGE_FORMATS.get(self.default_image_format.lower())

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info(self.emulator_details_text())

        self.__server = Server(self)
        self.__server.run()

        self.loaded = True

    def draw_to_screen(self, pixels):
        image = self._get_masked_image(pixels)
        with io.BytesIO() as bytesIO:
            image.save(
                bytesIO, self.image_format, quality=self.options.browser.quality, optimize=True
            )
            self.image = bytesIO.getvalue()




import tkinter
import os

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.graphics import Color
from RGBMatrixEmulator.logger import Logger


class TkinterAdapter(BaseAdapter):

    SUPPORTS_ALTERNATE_PIXEL_STYLE = True

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__root   = None
        self.__canvas = None
        self.__pixels = None

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info('Loading {}'.format(self.emulator_details_text()))
        self.__root = tkinter.Tk()
        self.__set_emulator_icon()
        self.__root.title(self.emulator_details_text())

        window_size = self.options.window_size()
        self.__root.geometry('{}x{}'.format(*window_size))
        self.__canvas = tkinter.Canvas(self.__root, 
                                       width=window_size[0], 
                                       height=window_size[1], 
                                       bd=0, 
                                       highlightthickness=0, 
                                       bg="black")

        self.__initialize_bitmap()
        self.__root.update()

        self.loaded = True

    def draw_to_screen(self, pixels):
        for row, pixel_row in enumerate(pixels):
            for col, pixel in enumerate(pixel_row):
                pixel = self.adjust_pixel_brightness(pixel)
                shape_id = self.__pixels[row][col]

                self.__canvas.itemconfig(shape_id, fill=Color.to_hex(pixel))

        self.__canvas.pack()
        self.__root.update()

    def __initialize_bitmap(self):
        self.__pixels = []

        for row in range(0, self.height):
            new_row = []

            for col in range(0, self.width):
                coords = self.__pixel_dimensions(col, row)

                if self.options.pixel_style == 'circle':
                    id = self.__canvas.create_oval(coords, width=0)
                else:
                    id = self.__canvas.create_rectangle(coords, width=0)

                new_row.append(id)

            self.__pixels.append(new_row)

    def __pixel_dimensions(self, col, row):
        size = self.options.pixel_size
        start, stop = (col * size, row * size)
            
        return (start, stop, start + size, stop + size)

    def __set_emulator_icon(self):
        emulator_path = os.path.abspath(os.path.dirname(__file__))
        raw_icon_path = os.path.join(emulator_path, '..', 'icon.ico')
        icon_path = os.path.normpath(raw_icon_path)

        self.__root.iconbitmap(icon_path)


import numpy as np
from PIL import Image
from RGBMatrixEmulator.graphics.color import Color

class Canvas:
    def __init__(self, options):
        self.options = options

        self.width = options.cols * options.chain_length
        self.height = options.rows * options.parallel
        self.display_adapter = options.display_adapter.get_instance(self.width, self.height, options)

        self.__pixels = [[Color.BLACK() for x in range(0, self.width)] for y in range(0, self.height)]

        self.display_adapter.load_emulator_window()

    def Clear(self):
        self.__pixels = [[Color.BLACK() for x in range(0, self.width)] for y in range(0, self.height)]

    def Fill(self, r, g, b):
        self.__pixels = [[(r, g, b) for x in range(0, self.width)] for y in range(0, self.height)]

    def SetPixel(self, x, y, r, g, b):
        if self.display_adapter.pixel_out_of_bounds(x, y):
            return

        pixel = self.__pixels[int(y)][int(x)] = (r, g, b)

    def SetImage(self, image, offset_x=0, offset_y=0, *other):
        original = Image.fromarray(np.array(self.__pixels, dtype=np.uint8), "RGB")
        original.paste(image, (offset_x, offset_y))
        self.__pixels = np.asarray(original)

    # These are delegated to the display adapter to handle specific implementation.
    def draw_to_screen(self):
        self.display_adapter.draw_to_screen(self.__pixels)

    def check_for_quit_event(self):
        self.display_adapter.check_for_quit_event()


from RGBMatrixEmulator.emulators.canvas import Canvas


class RGBMatrix:
    def __init__(self, options = {}):
        self.options = options

        self.width = options.cols * options.chain_length
        self.height = options.rows * options.parallel
        self.brightness = options.brightness

        self.canvas = None

    def CreateFrameCanvas(self):
        self.canvas = Canvas(options = self.options)

        return self.canvas
    
    def SwapOnVSync(self, canvas):
        canvas.check_for_quit_event()
        canvas.draw_to_screen()
        self.canvas = canvas

        return self.canvas

    def Clear(self):
        self.__sync_canvas()
        self.canvas.Clear()
        self.SwapOnVSync(self.canvas)

    def Fill(self, r, g, b):
        self.__sync_canvas()
        self.canvas.Fill(r, g, b)
        self.SwapOnVSync(self.canvas)

    def SetPixel(self, x, y, r, g, b):
        self.__sync_canvas()
        self.canvas.SetPixel(x, y, r, g, b)
        self.SwapOnVSync(self.canvas)

    def SetImage(self, image, offset_x=0, offset_y=0, *other):
        self.__sync_canvas()
        self.canvas.SetImage(image, offset_x, offset_y, *other)
        self.SwapOnVSync(self.canvas)
 
    def __sync_canvas(self):        
        if not self.canvas:
            self.canvas = Canvas(options = self.options)

        self.canvas.display_adapter.options.brightness = self.brightness


import os
import sys

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.graphics import Color


class TerminalAdapter(BaseAdapter):

    SUPPORTS_ALTERNATE_PIXEL_STYLE = True
    SYMBOLS = {
        "circle": " ●",
        "square": "██"
    }

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__symbol = self.SYMBOLS.get(self.options.pixel_style)

    def draw_to_screen(self, pixels):
        output = "\033[H\n"                                               # Move the cursor to the home position, add a little border
        for pixel_row in pixels:
            output += "  "                                                # Add a bit of border in case cursor causes line to wrap
            for pixel in pixel_row:
                pixel = self.adjust_pixel_brightness(pixel)
                output += "\033[38;2;{};{};{}m".format(*pixel) # Set the cell to the pixel color
                output += self.__symbol                                   # Draw the pixel
                output += "\033[37m"                                      # Reset the color

            output += " \n"

        sys.stdout.write(output)

    def load_emulator_window(self):
        os.system('cls||clear')
        os.system('mode con: cols={} lines={}'.format(self.width * 2 + 5, self.height + 3))


import json, logging

# Try to load the config from file. (Default: INFO)
try:
    with open("emulator_config.json") as config_file:
        log_level_name = json.load(config_file).get("log_level", 'INFO').upper()
        log_level = getattr(logging, log_level_name)
except:
    log_level = logging.INFO

# Create a Logger
Logger = logging.getLogger('RGBME')
Logger.setLevel(log_level)

# Create console handler and set the log level
ch = logging.StreamHandler()
ch.setLevel(log_level)

# Create formatter
formatter = logging.Formatter(
    '[%(asctime)s] [%(name)s] [%(levelname)s]: %(message)s',
    datefmt = '%Y-%m-%d %H:%M:%S'
)

# Add formatter to console handler
ch.setFormatter(formatter)

# Add console handler to Logger
Logger.addHandler(ch)



import bdfparser


class Font:
    def __init__(self):
        self.bdf_font = None
        self.headers = {}
        self.spacing = {}

    def LoadFont(self, path):
        self.bdf_font = bdfparser.Font(path)
        self.headers = self.bdf_font.headers
        self.props = self.bdf_font.props

        # All rpi-rgb-led-matrix fonts have a character at 0xFFFD to represent a missing character
        # Cache this for use later so we don't have to constantly look it up
        self.default_character = self.bdf_font.glyphbycp(0xFFFD)

    def CharacterWidth(self, char):
        # Missing glyphs return 0 width in rpi-rgb-led-matrix
        if self.bdf_font == None or not self.bdf_font.glyphbycp(char):
            return 0

        return self.bdf_font.glyphbycp(char).meta['dwx0']

    @property
    def height(self):
        if self.bdf_font is None: return -1
        return self.headers['fbby']

    @property
    def baseline(self):
        if self.bdf_font is None: return 0
        return self.headers['fbby'] + self.headers['fbbyoff']



import json, os, pprint, sys

from RGBMatrixEmulator.adapters import ADAPTER_TYPES
from RGBMatrixEmulator.logger import Logger


class RGBMatrixOptions:
    def __init__(self):
        self.hardware_mapping         = 'EMULATED'
        self.rows                     = 32
        self.cols                     = 32
        self.chain_length             = 1
        self.parallel                 = 1
        self.row_address_type         = 0
        self.multiplexing             = 0
        self.pwm_bits                 = 0
        self.brightness               = 100
        self.pwm_lsb_nanoseconds      = 130
        self.led_rgb_sequence         = 'RGB-EMULATED'
        self.show_refresh_rate        = 0
        self.gpio_slowdown            = None
        self.disable_hardware_pulsing = False

        emulator_config = RGBMatrixEmulatorConfig()

        if emulator_config.display_adapter.lower() in ADAPTER_TYPES:
            self.display_adapter = ADAPTER_TYPES[emulator_config.display_adapter.lower()]
        elif len(ADAPTER_TYPES.keys()) > 0:
            adapter_types = ', '.join('"{}"'.format(key) for key in ADAPTER_TYPES.keys())

            # Try to set it to the emulator default, but if it failed to load, pick the first one that did.
            if emulator_config.DEFAULT_CONFIG.get('display_adapter') in ADAPTER_TYPES:
                default_adapter = emulator_config.DEFAULT_CONFIG.get('display_adapter')
            else:
                default_adapter = list(ADAPTER_TYPES.keys())[0]

            Logger.warning('"{}" display adapter option not recognized. Valid adapters are {}. Defaulting to "{}"...'.format(
                    emulator_config.display_adapter,
                    adapter_types,
                    default_adapter
                )
            )
            self.display_adapter = ADAPTER_TYPES[default_adapter]
        else:
            Logger.critical("Failed to find a valid display adapter to load! Check that you have installed dependencies required for your configured adapter.")

            sys.exit(1)

        self.pixel_style = emulator_config.DEFAULT_CONFIG.get('pixel_style')
        config_pixel_style = emulator_config.pixel_style.lower()

        if config_pixel_style in emulator_config.VALID_PIXEL_STYLES:
            if config_pixel_style != self.pixel_style:
                if self.display_adapter.SUPPORTS_ALTERNATE_PIXEL_STYLE:
                    self.pixel_style = emulator_config.pixel_style
                else:
                    Logger.warning('"{}" pixel style option is not supported by adapter "{}". Defaulting to "square"...'.format(config_pixel_style, emulator_config.display_adapter.lower()))
        else:
            Logger.warning('"{}" pixel style option not recognized. Valid options are "square", "circle". Defaulting to "square"...'.format(config_pixel_style))

        self.pixel_size = emulator_config.pixel_size
        self.pixel_outline = emulator_config.DEFAULT_CONFIG['pixel_outline']
        self.pixel_outline = emulator_config.pixel_outline
        self.browser = emulator_config.browser

        if emulator_config.suppress_font_warnings:
            import bdfparser

            bdfparser.warnings.simplefilter("ignore")

    def window_size(self):
        return (self.cols * self.pixel_size * self.chain_length, self.rows * self.pixel_size * self.parallel)

    def window_size_str(self, pixel_text=""):
        width, height = self.window_size()

        return f"{width} x {height} {pixel_text}"

class RGBMatrixEmulatorConfig:

    __CONFIG_PATH = 'emulator_config.json'

    VALID_PIXEL_STYLES = ['square', 'circle']
    DEFAULT_CONFIG = {
        'pixel_outline': 0,
        'pixel_size': 16,
        'pixel_style': 'square',
        'display_adapter': 'browser',
        'suppress_font_warnings': False,
        'suppress_adapter_load_errors': False,
        'browser': {
            '_comment': 'For use with the browser adapter only.',
            'port': 8888,
            'target_fps': 24,
            'fps_display': False,
            'quality': 70,
            'image_border': True,
            'debug_text': False,
            'image_format': 'JPEG'
        },
        "log_level": "info"
    }

    def __init__(self):
        self.config         = self.__load_config()
        self.default_config = self.DEFAULT_CONFIG

        RGBMatrixEmulatorConfig.Utils.set_attributes(self)

    def __load_config(self):
        if os.path.exists(self.__CONFIG_PATH):
            with open(self.__CONFIG_PATH) as f:
                config = json.load(f)

            return config

        with open(self.__CONFIG_PATH, 'w') as f:
            json.dump(self.DEFAULT_CONFIG, f, indent=4)

        return self.DEFAULT_CONFIG

    def __str__(self):
        return RGBMatrixEmulatorConfig.Utils.to_str(self)


    class ChildConfig:
        def __init__(self, config, default_config):
            self.config         = config
            self.default_config = default_config

            RGBMatrixEmulatorConfig.Utils.set_attributes(self)

        def __str__(self):
            return RGBMatrixEmulatorConfig.Utils.to_str(self)


    class Utils:
        def to_str(obj):
            '''
            Pretty prints the config object from dict.
            '''
            printer = pprint.PrettyPrinter(sort_dicts=False)
            return "\n".join([obj.__repr__(), printer.pformat(RGBMatrixEmulatorConfig.Utils.to_dict(obj))])

        def to_dict(obj):
            '''
            Recursively recreates the config dict from child config objects.
            '''
            config = {}
            for key in obj.__dict__.keys():
                if key in ['config', 'default_config'] or key[0] == '_':
                    continue

                value = obj.__dict__.get(key)

                if isinstance(value, RGBMatrixEmulatorConfig.ChildConfig):
                    value = RGBMatrixEmulatorConfig.Utils.to_dict(value)

                config[key] = value

            return config

        def set_attributes(obj):
            '''
            Dynamically set attributes loaded into config and default config variables.

            Numbers, strings, and arrays are stored natively. Nested dicts are parsed into RGBMatrixEmulatorChildConfig objects recursively.
            '''
            for key in obj.default_config.keys():
                if key in obj.config:
                    value   = obj.config.get(key)
                    default = obj.default_config.get(key)
                else:
                    value   = obj.default_config.get(key)
                    default = value

                    Logger.warning("Emulator config is missing key '{}', falling back to default '{}'. Consider adding this to your emulator config file.".format(key, value))

                RGBMatrixEmulatorConfig.Utils.set_attribute(obj, key, value, default)

        def set_attribute(obj, key, value, default):
            '''
            Store the value as an attribute or delegate to the RGBMatrixEmulatorChildConfig to parse into a new node.
            '''
            if isinstance(value, dict):
                obj.__setattr__(key, RGBMatrixEmulatorConfig.ChildConfig(value, default))
            else:
                obj.__setattr__(key, value)



import tornado.web


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.render('./../static/index.html', adapter=MainHandler.adapter)

    def register_adapter(adapter):
        MainHandler.adapter = adapter



import numpy as np

from PIL import Image, ImageDraw, ImageEnhance
from RGBMatrixEmulator import version
from RGBMatrixEmulator.graphics import Color

def draw_circle_mask(drawer, x, y, pixel_size, color):
    drawer.ellipse(
        (x, y, x + pixel_size - 1, y + pixel_size - 1),
        fill=color,
        outline=color,
    )

def draw_square_mask(drawer, x, y, pixel_size, color):
    drawer.rectangle(
        (x, y, x + pixel_size, y + pixel_size),
        fill=color,
        outline=color,
        )

class BaseAdapter:

    SUPPORTS_ALTERNATE_PIXEL_STYLE = False
    INSTANCE = None

    def __init__(self, width, height, options):
        self.width   = width
        self.height  = height
        self.options = options
        self.__black = Image.new("RGB", self.options.window_size(), "black")
        self.__mask = self.__draw_mask()
        self.loaded  = False

    def __draw_mask(self):
        mask = Image.new("L", self.options.window_size())
        drawer = ImageDraw.Draw(mask)
        pixel_size = self.options.pixel_size
        width, height = self.options.window_size()
        color = int((self.options.brightness * 255) / 100)
        draw_mask_shape = (draw_circle_mask
            if self.options.pixel_style == "circle" else draw_square_mask)
        for y in range(0, height, pixel_size):
            for x in range(0, width, pixel_size):
                draw_mask_shape(drawer, x, y, pixel_size, color)

        return mask

    @classmethod
    def get_instance(cls, *args, **kwargs):
        if cls.INSTANCE is None:
            instance = cls(*args, **kwargs)
            cls.INSTANCE = instance
        
        return cls.INSTANCE

    def adjust_pixel_brightness(self, pixel, to_int = True):
        alpha = self.options.brightness / 100.0
        return Color.adjust_brightness(pixel, alpha, to_int = to_int)

    def pixel_out_of_bounds(self, x, y):
        if x < 0 or x >= self.width:
            return True

        if y < 0 or y >= self.height:
            return True

        return False

    def _get_masked_image(self, pixels):
        image = Image.fromarray(np.array(pixels, dtype=np.uint8), "RGB")
        image = image.resize(self.options.window_size(), Image.NEAREST)
        enhancer = ImageEnhance.Brightness(image)
        image = enhancer.enhance(self.options.brightness / 100.0)

        return Image.composite(image, self.__black, self.__mask)

    def emulator_details_text(self):
        details_text = 'RGBME v{} - {}x{} Matrix | {}x{} Chain | {}px per LED ({}) | {}'

        return details_text.format(version.__version__,
                                   self.options.cols,
                                   self.options.rows,
                                   self.options.chain_length,
                                   self.options.parallel,
                                   self.options.pixel_size,
                                   self.options.pixel_style.upper(),
                                   self.__class__.__name__)

    # This method is required for the pygame adapter but nothing else, so just skip it if not defined.
    def check_for_quit_event(self):
        pass

    #############################################################
    # These methods must be implemented by BaseAdapter subclasses
    #############################################################
    def load_emulator_window(self):
        '''
        Initialize the external dependency as a graphics display.

        This method is fired when the emulated canvas is initialized.
        '''
        raise NotImplementedError

    def draw_to_screen(self, _pixels):
        '''
        Accepts a 2D array of pixels of size height x width.

        Implements drawing each pixel to the screen via the external dependency loaded in load_emulator_window.
        Before drawing, use adjust_pixel_brightness() on each pixel if your display adapter supports it.
        '''
        raise NotImplementedError



class Color:
    def __init__(self, r = 0, g = 0, b = 0):
        self.red   = r
        self.green = g
        self.blue  = b

    @classmethod
    def adjust_brightness(cls, pixel, alpha, to_int = False):
        p = tuple(channel * alpha for channel in pixel)

        if to_int:
            p = tuple(int(channel) for channel in pixel)

        return p

    @classmethod
    def to_hex(cls, pixel):
        return "#%02x%02x%02x" % pixel

    @classmethod
    def BLACK(cls):
        return (0, 0, 0)

    @classmethod
    def RED(cls):
        return (255, 0, 0)

    @classmethod
    def GREEN(cls):
        return (0, 255, 0)

    @classmethod
    def BLUE(cls):
        return (0, 0, 255)





import os
import platform
import sys
import io
import libsixel as sixel

from PIL import Image, ImageDraw, ImageEnhance
from typing import List

from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.graphics.color import Color


class SixelAdapter(BaseAdapter):

    SUPPORTS_ALTERNATE_PIXEL_STYLE = True

    def draw_to_screen(self, pixels):
        sixel = self.__encode_sixels(pixels,
                                     scale=self.options.pixel_size,
                                     outline=self.options.pixel_outline)
        output = f"\033[H\n{sixel}\n"
        sys.stdout.write(output)

    def load_emulator_window(self):
        os.system('cls||clear')
        if platform.system() == "Windows":
            os.system('mode con: cols={} lines={}'.format(self.width * 2 + 5, self.height + 3))
        else:
            os.system('stty columns {} rows {}'.format(self.width * 2 + 5, self.height + 3))

    def __enlarge_pixels(self, pixels: List[List[Color]], scale: int, outline: int) -> Image.Image:
        outline_color = '#000000'
        w, h = len(pixels[0]), len(pixels)
        iw, ih = w * scale, h * scale
        i = Image.new('RGBA', (iw, ih))
        d = ImageDraw.Draw(i)
        for y, row in enumerate(pixels):
            for x, pixel in enumerate(row):
                rx, ry = x * scale, y * scale
                ex, ey = rx + (scale - 1), ry + (scale - 1)
                pixel = self.adjust_pixel_brightness(pixel)
                if self.options.pixel_style == "circle":
                    d.ellipse([(rx, ry), (ex, ey)], fill=pixel, outline=outline_color, width=outline)
                else:
                    d.rectangle([(rx, ry), (ex, ey)], fill=pixel, outline=outline_color, width=outline)

        brightness = ImageEnhance.Brightness(i)
        i = brightness.enhance(1.5)
        contrast = ImageEnhance.Contrast(i)
        i = contrast.enhance(0.8)

        return i

    def __encode_sixels(self, pixels: List[List[Color]], scale=4, outline=1) -> str:
        '''Encodes given Image to a sixel string.'''
        img = self.__enlarge_pixels(pixels, scale, outline)
        img_data = img.convert('RGB').tobytes()
        width = img.width
        height = img.height
        with io.BytesIO() as buf:
            output = sixel.sixel_output_new(lambda data, buffer: buffer.write(data), buf)
            dither = sixel.sixel_dither_new(256)
            sixel.sixel_dither_initialize(dither, img_data, width, height, sixel.SIXEL_PIXELFORMAT_RGB888)
            sixel.sixel_encode(img_data, width, height, 1, dither, output)

            encoded = buf.getvalue().decode('ascii')

            sixel.sixel_dither_unref(dither)
            sixel.sixel_output_unref(output)

        return encoded



#!/usr/bin/env python

from setuptools import find_packages, setup
import codecs
import os.path as path

cwd = path.dirname(__file__)

with open("description.md", "r") as fh:
    long_description = fh.read()

version = '0.0.0'
with codecs.open(path.join(cwd, 'RGBMatrixEmulator/version.py'), 'r', 'ascii') as f:
    exec(f.read())
    version = __version__
assert version != '0.0.0'

setup(
    name='RGBMatrixEmulator',
    author='Tyler Porter',
    author_email='tyler.b.porter@gmail.com',
    version=version,
    license='MIT',
    description='A PC emulator for Raspberry Pi LED matrices driven by rpi-rgb-led-matrix',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='https://github.com/ty-porter/RGBMatrixEmulator',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Natural Language :: English',
        'Topic :: Other/Nonlisted Topic',
    ],
    keywords=[
        'LED matrix',
        'matrix',
        'raspberry pi',
        'raspberry',
        'RPI',
        'LED'
    ],
    platforms='ANY',
    packages=find_packages(),
    data_files=[
        ('docs', ['README.md', 'LICENSE', 'description.md']),
        ('RGBMatrixEmulator', [
            'RGBMatrixEmulator/icon.ico',
            'RGBMatrixEmulator/icon.png',
            'RGBMatrixEmulator/adapters/browser_adapter/static/index.html',
            'RGBMatrixEmulator/adapters/browser_adapter/static/assets/client.js',
            'RGBMatrixEmulator/adapters/browser_adapter/static/assets/icon.ico',
            'RGBMatrixEmulator/adapters/browser_adapter/static/assets/styles.css',
        ])
    ],
    install_requires=[
        'bdfparser<=2.2.0',
        'numpy>=1.2.0',
        'pygame>=2.0.1,<3',
        'tornado>=6.1',
        'libsixel-python>=0.5.0',
        'Pillow>=10.0.1',
    ],
    include_package_data=True
)



import asyncio
import signal
import sys
import threading
import tornado.web
import tornado.ioloop

from os import path
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from RGBMatrixEmulator.adapters.browser_adapter.request_handlers import *
from RGBMatrixEmulator.logger import Logger


asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

class Server:

    instance = None

    # Singleton class for the server.
    # No more than one webserver can be running at a given time.
    class __Singleton:

        def __init__(self, adapter):
            self.adapter   = adapter
            self.io_loop   = None
            self.listening = False

            MainHandler.register_adapter(self.adapter)
            ImageWebSocketHandler.register_adapter(self.adapter)
            ImageHandler.register_adapter(self.adapter)

            script_path  = path.dirname(path.realpath(__file__))
            asset_path   = path.normpath(script_path + '/static/assets/')

            self.app = tornado.web.Application([
                (r"/websocket", ImageWebSocketHandler),
                (r"/image", ImageHandler),
                (r"/", MainHandler),
                (r"/assets/(.*)", tornado.web.StaticFileHandler, { 'path': asset_path, 'default_filename': 'client.js' })
            ])

    def __init__(self, adapter):
        if not Server.instance:
            Server.instance = Server.__Singleton(adapter)

    def run(self):     
        if not self.instance.listening:
            Logger.info("Starting server...")

            self.instance.listening = True
            self.instance.app.listen(self.instance.adapter.options.browser.port)
            self.instance.io_loop = tornado.ioloop.IOLoop.current()
            thread = threading.Thread(target=self.instance.io_loop.start, name="RGBMEServerThread", daemon=True)
            self.__initialize_interrupts()
            thread.start()

            Logger.info("Server started and ready to accept requests on http://localhost:" + str(self.instance.adapter.options.browser.port) + "/")

    def __initialize_interrupts(self):
        '''
        Add custom signal handling to ensure webserver thread exits appropriately.

        Not thread-safe, signal handling must happen on the main thread.
        '''
        if threading.current_thread() is threading.main_thread():
            signal.signal(signal.SIGINT , self.__kill)
            signal.signal(signal.SIGTERM, self.__kill)

    def __kill(self, *_args):
        self.instance.io_loop.add_callback(self.instance.io_loop.stop)
        sys.exit()



import numpy as np
import os
import sys

# Try to suppress the pygame load warning if able.
try:
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
except Exception:
    pass

import pygame

from PIL import Image
from pygame.locals import QUIT
from RGBMatrixEmulator.adapters.base import BaseAdapter
from RGBMatrixEmulator.graphics import Color
from RGBMatrixEmulator.logger import Logger


class PygameAdapter(BaseAdapter):

    SUPPORTS_ALTERNATE_PIXEL_STYLE = True

    def __init__(self, width, height, options):
        super().__init__(width, height, options)
        self.__surface = None

    def load_emulator_window(self):
        if self.loaded:
            return

        Logger.info('Loading {}'.format(self.emulator_details_text()))
        self.__surface = pygame.display.set_mode(self.options.window_size())
        pygame.init()

        self.__set_emulator_icon()
        pygame.display.set_caption(self.emulator_details_text())

        self.loaded = True

    def draw_to_screen(self, pixels):
        image = self._get_masked_image(pixels)
        pygame_surface = pygame.image.fromstring(image.tobytes(),
            self.options.window_size(), "RGB")
        self.__surface.blit(pygame_surface, (0, 0))

        pygame.display.flip()

    def check_for_quit_event(self):
        # We don't have events, but this will keep the emulator from appearing as if it's not responding.
        # This also enables closing the window to kill the emulator
        for event in pygame.event.get():
            if event.type == QUIT:
                pygame.quit()
                sys.exit()

    def __set_emulator_icon(self):
        emulator_path = os.path.abspath(os.path.dirname(__file__))
        icon_path = os.path.join(emulator_path, '..', 'icon.png')
        icon = pygame.image.load(os.path.normpath(icon_path))

        pygame.display.set_icon(icon)

    def __pygame_pixel(self, col, row):
        return pygame.Rect(
            col * self.options.pixel_size,
            row * self.options.pixel_size,
            self.options.pixel_size,
            self.options.pixel_size
        )




import tornado.websocket

from RGBMatrixEmulator.logger import Logger


class ImageWebSocketHandler(tornado.websocket.WebSocketHandler):
    clients = set()
    adapter = None

    def check_origin(self, _origin):
        # Allow access from every origin
        return True

    def open(self):
        ImageWebSocketHandler.clients.add(self)
        Logger.info("WebSocket opened from: " + self.request.remote_ip)

    def on_message(self, _message):
        if not ImageWebSocketHandler.adapter.image:
            Logger.warning("No image received from {}!".format(ImageWebSocketHandler.adapter.__class__.__name__))
            return

        image_bytes = ImageWebSocketHandler.adapter.image
        self.write_message(image_bytes, binary=True)

    def on_close(self):
        ImageWebSocketHandler.clients.remove(self)

    def register_adapter(adapter):
        ImageWebSocketHandler.adapter = adapter
